// 开启单词可点击
var enableWebWordClickable = true;
var enableOpenExternalUrl = true;
var isAsyncDomLoaded = false ;
var debugMode = false;

if (typeof window.levelWordSet === "undefined") {
    window.levelWordSet = new Set();
}

if (typeof window.currentLevel === "undefined") {
    window.currentLevel = "";
}

function setDebugMode(debug) {
    debugMode = debug;
}

function isInvalidNode(node) {
    if (!node) {
        return true
    }

    switch (node.nodeType) {
        case Node.TEXT_NODE:
            return false
        case Node.COMMENT_NODE:
        case Node.CDATA_SECTION_NODE:
            return true
        case Node.ELEMENT_NODE:
            {
                // 不包含文本内容的标签
                switch (node.tagName) {
                    case 'IMG':
                    case 'SVG':
                    case 'VIDEO':
                    case 'CANVAS':
                    case 'SCRIPT':
                    case 'NOSCRIPT':
                    case 'COMMENT':
                    case 'CDATA':
                    case 'AUDIO':
                    case 'INPUT':
                    case 'TRACK':
                    case 'PATH':
                        return true
                }
            }
    }
    return false
}

function textNodesUnder(node) {
    var all = [];
    if (debugMode) {
        console.log('node tag: ' + node.tagName);
    }
    if (isInvalidNode(node)) {
        if (debugMode) {
            console.log('>>>>>> SKIP tag: ' + node.tagName);
        }
    } else {
        for (node = node.firstChild; node; node = node.nextSibling) {
            if (node.nodeType == Node.TEXT_NODE) {
                all.push(node);
            } else if (node.nodeType == Node.ELEMENT_NODE) {
                all = all.concat(textNodesUnder(node));
            }
        }
    }
    return all;
}


function createSelectionRange(elm) {
    var range = document.createRange();
    range.setStart(elm, 0);
    range.setEnd(elm, 1);
    window.getSelection().removeAllRanges();
    window.getSelection().addRange(range);
}

function removeSelectionRanges() {
    console.log("-> removeSelectionRanges");
    window.getSelection().removeAllRanges();
}

function isChildOfHyperLink(elem) {
    var currentElem = elem;
    while ( currentElem.parentNode ) {
//        console.log("currentElem.parentNode = " + currentElem.parentNode.tagName);
        if (currentElem.tagName === "A" || currentElem.parentNode.tagName === "A") {
            console.log("isChildOfHyperLink = true");
            return true;
        }
        currentElem = currentElem.parentNode;
    }
    return false;
}

function clickWordSpan(elem, wordLevel) {
    if (!enableWebWordClickable) {
        console.log('word can not clicked.');
        return;
    }

    // 停止事件传递
    event.stopPropagation();
    var word = elem.innerText.trim();
    // 字母、'、’、横杠为有效字符
    let wordParts = word.match(/[a-zA-Z'-/’]+/g);
    if (wordParts != null && wordParts.length > 0) {
        word = wordParts[0];
        console.log("word from regex: " + word);
        // 选中单击的元素
        createSelectionRange(elem);
    } else {
        console.log(word + ' is not word .');
        return;
    }

    // click hyper link child node
    if (isChildOfHyperLink(elem)) {
        if (!enableOpenExternalUrl) {
            // 阻止默认行为，即阻止跳转
            event.preventDefault();
            // 在此处可以添加其他自定义逻辑
            console.log("超链接点击事件被禁止");
        } else {
            console.log("超链接内部tag被点击 => url 跳转");
            return;
        }
    }

    word = word.replace(".", "").replace("?", "").replace("!", "").replace(",", "").replace("\"", "")
    if (word.endsWith("'") || word.endsWith("’")) {
        word = word.substring(0, word.length - 1)
    }
    console.log('clickWordSpan : ' + word);

    let rect = getRectForSelectedText(elem);
    if (rect != null) {
        let selectWordSentence = getSelectionSentence();
        if (typeof FolioWebView != "undefined" && typeof eval(FolioWebView.onClickWordToExplain) === "function") {
            // 点击普通单词
            FolioWebView.onClickWordToExplain(rect.left, rect.top, rect.right, rect.bottom, word, selectWordSentence, wordLevel);
        } else {
            console.log('word = ' + word + ", rect: " + JSON.stringify(rect) + ", sentence : " + selectWordSentence);
        }
    }
}


// Method that gets the Rect of current selected text
// and returns in a JSON format
var getRectForSelectedText = function(elm) {
    if (typeof elm === "undefined") {
        elm = window.getSelection().getRangeAt(0);
    }
    return /*elm.getBoundingClientRect()*/ RangeFix.getBoundingClientRect(elm);
};


function wrapTextNodeToWordSpan(textNode) {
    let nodeText = textNode.textContent;
    // console.log('node text : ' + nodeText);

    if (nodeText == null || nodeText.trim().length == 0) {
        return
    }

    // 通过空格拆分单词, 并且只提取英文单词
    let words = nodeText.split(/\s+/).filter(word => /[0-9a-zA-Z'-/’\\.,?!]+/.test(word));
    if (words != null && typeof words !== "undefined") {
        var innerHTML = ""

        let wrapSpanElem = document.createElement('span');
        words.forEach(function(word, index) {
//            console.log('wrap word: ' + word);
            if (word.trim().length > 0) {
                let wordSpanElem = document.createElement('span');
                var wordLevel = "";
                if (typeof window.levelWordSet !== "undefined") {
                    wordLevel = window.levelWordSet.has(word.toLowerCase()) ? window.currentLevel : "";
                }

                let wordSpanClass = wordLevel.trim().length > 0 ? "level_word_" + wordLevel : "";
                if (wordSpanClass.length > 0) {
                    wordSpanElem.setAttribute('class', wordSpanClass);
                }

                wordSpanElem.addEventListener('click', function handleClick(event) {
                    clickWordSpan(wordSpanElem, wordLevel);
                });
                wordSpanElem.appendChild(document.createTextNode(word));

                // 追加到最外层的wrapSpanElem中
                wrapSpanElem.appendChild(wordSpanElem);
                wrapSpanElem.appendChild(document.createTextNode(" "));
            } else {
                // 追加到最外层的wrapSpanElem中
                wrapSpanElem.appendChild(document.createTextNode(word));
            }
        });

        if (textNode.parentNode != null) {
            textNode.parentNode.replaceChild(wrapSpanElem, textNode);
        }
    }
}

function rebuildHtml() {
    var body = document.getElementsByTagName("body")[0];
    var textNodes = textNodesUnder(body);
    textNodes.forEach(function(element, index) {
        if (element.nodeValue != null && element.nodeValue.toString().length > 0) {
            // console.log('node : ' + index + ", text = " + element.nodeValue);
            wrapTextNodeToWordSpan(element);
        } else {
            // console.log('=======>>> empty node : ' + index);
        }
    });

    // 回调客户端
    if (typeof FolioWebView != "undefined" && typeof eval(FolioWebView.onClickWordToExplain) === "function") {
        FolioWebView.onWebRebuildFinish(textNodes.length > 0);
        console.log('=======>>> onWebRebuildFinish');
    }
    console.log('=======>>> rebuildHtml done!');
}


function setWebWordClickable(enable) {
    console.log('setWebWordClickable : ' + enable);
    enableWebWordClickable = enable;
}


function setEnableOpenExternalUrl(enable) {
   enableOpenExternalUrl = enable;
   console.log("setEnableOpenExternalUrl : " + enable);
}

function setupHyperLinkClicks() {
    // 获取所有超链接 <a> 元素
    var links = document.querySelectorAll("a");
    if (links && links.length > 0) {
        // 遍历所有链接并为它们添加点击事件监听器
        links.forEach(function(link) {
            link.addEventListener("click", function(event) {
                if (!enableOpenExternalUrl) {
                    // 阻止默认行为，即阻止跳转
                    event.preventDefault();
                    // 在此处可以添加其他自定义逻辑
                    console.log("超链接点击事件被禁止");
                } else {
                    console.log("超链接被点击");
                }
            });
        });
    }
}

function rebuildAsyncDOMHtml() {
    if ( isAsyncDomLoaded ) {
        setWebWordClickable(true);
        return;
    }
    rebuildHtml();
    setupHyperLinkClicks();
    setWebWordClickable(true);
    isAsyncDomLoaded = true;
}

console.log("maybe document load done => rebuildHtml");
rebuildHtml();
setupHyperLinkClicks();


//// 使用 JavaScript 添加点击事件监听器
//document.addEventListener("DOMContentLoaded", function() {
//    setupHyperLinkClicks();
//});